package at.tugraz.iaik.p2p.evt.msg

import at.tugraz.iaik.p2p.commons.readByteArray
import at.tugraz.iaik.p2p.commons.writeByteArray
import at.tugraz.iaik.p2p.commons.writeShort
import at.tugraz.iaik.p2p.pol.*
import at.tugraz.iaik.p2p.skad.id.ECID
import at.tugraz.iaik.p2p.skad.id.readIdentifier
import at.tugraz.iaik.p2p.skad.id.writeIdentifier
import at.tugraz.iaik.p2p.skad.msg.ExtendedMessage
import at.tugraz.iaik.p2p.skad.msg.ExtendedType
import at.tugraz.iaik.p2p.skad.msg.Message
import at.tugraz.iaik.p2p.skad.msg.MessageHeader
import at.tugraz.iaik.p2p.skad.node.Node
import java.io.DataInput
import java.io.DataOutput

/**
 * Custom messages for policy evaluation in top of kademlia base p2p network see message base classes for more information
 */
enum class Error {
    UNKNOWN_MSG, EVL_MISMATCH, INVALID_SHARE_SIG, INVALID_POLICY_SIG, UNAUTHORIZED
}

//Identifiers for custom message types; use some random long as base
private var offset = 645710602390000L
val T_ERR = ++offset
val T_PEV = ++offset
val T_PER = ++offset

abstract class PolRequestMessage(
    extendedType: ExtendedType,
    val acceptedType: ExtendedType,
    seqId: Long,
    origin: Node,
    deser: (header: MessageHeader, buf: DataInput) -> Message
) : ExtendedMessage(extendedType, seqId, origin, deser)

class ErrorMessage(seqId: Long, origin: Node, val code: Error, val msg: String) :
    ExtendedMessage(
        T_ERR,
        seqId,
        origin,
        deser = { rd, buffer ->
            ErrorMessage(
                rd.seqId,
                rd.origin,
                Error.values()[buffer.readShort().toUShort().toInt()],
                String(buffer.readByteArray())
            )
        }) {
    override fun serialize(buffer: DataOutput) {

        buffer.writeShort(code.ordinal.toUShort().toShort())
        buffer.writeByteArray(msg.toByteArray())
    }

    override fun toString(): String {
        return "ErrorMessage(seqId=$seqId, origin=$origin, code=$code, msg=\"$msg\")"
    }
}

class EvalRequestMessage(seqId: Long, origin: Node, val share: Share, val policy: SignedPolicy, val creator: ECID) :
    PolRequestMessage(
        T_PEV, T_PER,
        seqId,
        origin,
        deser = { rd, buffer ->
            EvalRequestMessage(
                rd.seqId,
                rd.origin,
                buffer.readShare(),
                buffer.readSignedPolicy(),
                buffer.readIdentifier() as ECID
            )
        }) {
    override fun serialize(buffer: DataOutput) {
        buffer.writeShare(share)
        buffer.writeSignedPolicy(policy)
        buffer.writeIdentifier(creator)
    }

}

class EvalResponseMessage(seqId: Long, origin: Node, val recoveredShare: ByteArray) :
    ExtendedMessage(
        T_PER,
        seqId,
        origin,
        deser = { rd, buffer ->
            EvalResponseMessage(
                rd.seqId,
                rd.origin,
                buffer.readByteArray()
            )
        }) {
    override fun serialize(buffer: DataOutput) {
        buffer.writeByteArray(recoveredShare)
    }
}