/*
 * This Kotlin source file was generated by the Gradle 'init' task.
 */
package at.tugraz.iaik.p2p.pol

import at.tugraz.iaik.p2p.commons.CommonCrypto
import at.tugraz.iaik.p2p.skad.config.Config
import at.tugraz.iaik.p2p.skad.id.ECID
import kotlinx.coroutines.Job
import kotlinx.coroutines.joinAll
import kotlinx.coroutines.launch
import kotlinx.coroutines.runBlocking
import kotlin.test.Test
import kotlin.test.assertTrue

class PolExample {
    @Test
    fun testAccessControl() {

        //generate EC KeyPair for creator
        val cKP = CommonCrypto.genKP(ed = false)

        //init creator
        val en1 = PolNode(Config(cKP, listenAddr = "localhost:1916"))

        //generate some random nodes and bootstrap to creator
        val nodes = mutableListOf<PolNode>()
        val jobs = mutableListOf<Job>()
        for (i in 1..20) {
            val conf2 = Config(CommonCrypto.genKP(false), listenAddr = "localhost:${1968 + i}")
            val en2 = PolNode(conf2).apply { nodes.add(this) }
            runBlocking {
                jobs.add(launch { en2.bootstrap(en1) })
            }
        }
        runBlocking { jobs.joinAll() }
        println("\n\n\nBootstrapping done\n\n\n")

        //define evaluators
        val evals = setOf(
            nodes[0].identifier as ECID,
            nodes[1].identifier as ECID,
            nodes[2].identifier as ECID,
            nodes[3].identifier as ECID,
            nodes[4].identifier as ECID,
            nodes[5].identifier as ECID
        )

        //define data to protect
        val plaintext = "My super secret plaintext"

        //define policy to be enforced (nodes 8,9,10 may access)
        //in reality, this will be read from a yaml file
        //policies use j-easy and MVEL
        val policy =
            setOf(nodes[8], nodes[9], nodes[10]).joinToString(separator = "\n") {
                """
                ---
                name: "accessor: ${it.identifier}"
                description: "${it.identifier} can access"
                
                condition: "sender.equals(\"${it.identifier}\")"
                actions:
                  - "accessGranted = true;"
                """.trimIndent()
            }

        //put all together into data container
        val container = SignedDataContainer.create(
            en1.identifier as ECID,
            cKP.private,
            plaintext.toByteArray(),
            IdBasedPolicy(policy),
            evals,
            3.toUByte()
        )

        runBlocking {
            //let node 9 try to gain access to data
            val recoveredMessage = String(nodes[9].decrypt(container))
            println("recovered Message: $recoveredMessage\n\n")
            assertTrue { recoveredMessage == plaintext }

            try {
                //now let node 11 (which is not defined in the policy) try to access the data
                String(nodes[11].decrypt(container))
            } catch (t: PolNode.NotEnoughSharesException) {
                assertTrue { t.present == 0 }
            }
        }
    }
}
